from typing import ParamSpec, TypeVar

T = TypeVar("T")
V = TypeVar("V")
P = ParamSpec("P")


PROTOCOL_NAME = "OneBot"
PROTOCOL_VERSION = "11"
MODULE_AUTHOR = "Meloland"
PROTOCOL_IDENTIFIER = f"{PROTOCOL_NAME}-v{PROTOCOL_VERSION}@{MODULE_AUTHOR}"
